import asyncio
import logging
import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from queue import Queue
from threading import Thread
from time import sleep
from typing import Iterable, Optional

import cv2
import numpy as np
from apscheduler.triggers.cron import CronTrigger
from data_access.visual_knowledge_db import SystemDB, UserDB
from pydantic.dataclasses import dataclass
from tzlocal import get_localzone

from ..utils import constants as const
from ..utils import feature_flags as ft
from ..utils import index as utils
from . import errors, func
from .utils import Detector, Streamer


class ImageGenerator(ABC):

    wait_message: Optional[str] = "wait"
    __images_queue: Queue = Queue()
    __wait_image: Optional[np.array] = None
    
    @abstractmethod
    def __aiter__(self) -> None:
        pass
    
    def set_wait_image(self) -> None:

        if self.__wait_image is None:
            self.__wait_image = np.zeros([h,w,1],dtype=np.uint8)
            center_image = (self.__wait_image.shape[0] // 2, self.__wait_image.shape[1] // 2)

        cv2.putText(image,"wait", center_image, cv2.CV_FONT_HERSHEY_SIMPLEX, 2, 255)

    
@dataclass
class VisualKnowledge(ImageGenerator):
    
    type_service: str
    db_properties: dict
    db_name: str
    max_descriptor_distance: float
    min_date_knowledge: float
    min_frequency: Optional[float] = 0.7
    resize_factor: Optional[float] = 0.25
    features: Optional[list] = [],
    type_system: Optional[str] = const.TYPE_SYSTEM["OPTIMIZED"],
    title: Optional[str] = str(uuid.uuid4())

    """
    Main model builder

    constructor
    :Parameters:
        {String} type_service - type of the listen server.
        {Dict} db - proprties of bd see https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html#pymongo.mongo_client.MongoClient.
        {String} db_name - name of db to be used.
        {Array} min_date_knowledge [min_date_knowledge=null] - minimum interval to determine a known user.
        {Number} min_frequency [min_frequency=0.7] - minimum frequency between days detectioned.
        {list} features [features=null] - characteristics that will be saved in each detection.
        {String} max_descriptor_distance [max_descriptor_distance=null] - max distance of diference between detections.
        {String} type_system [type_system=null] - type of system.
        {String} title [title=null] - title of system.

    :Returns:
        None.

    """
    def __post_init__(self):

        #properties
        self.id = None

        #agents 
        self.detector = None
        self.receiver = None
        self.streamer = None
        self.stream_fps = 30

        #more info
        self.type_model_detection = None
        self.display_size = (720, 720)
        self.matches = None
        self.interval_streaming = None
        self.execution = False
        self.date_format = const.DATE_FORMAT
        self.draw_label = True
        self.cron_reload = self.__get_cron_trigger()

        #DB
        self.db_systems = SystemDB(
            properties = self.db_properties,
            db = self.db_name,
            collection = const.COLLECTIONS["SYSTEMS"]
        )
        self.db_users = UserDB(
            properties = self.db_properties,
            db = self.db_name,
            collection = const.COLLECTIONS["USERS"]
        )
    
    def __aiter__(self) -> Iterable:
      return self
    
    async def __anext__(self) -> np.array:

        if self.__images_queue.empty():
            self.set_wait_image()
        
        image: np.array = await self.__images_queue.get()


    async def __preload(self) -> None:
        """
            Load system and users if exist

            Search the system or saved it
            Search the users of the system
        """
        # found or add system
        system = await func.get_system(self.get_obj(), self.db_systems)
        if system is None:
            system = await func.insert_system(self.get_obj(), self.db_systems)
            if system is None:
                raise errors.SystemNotFoundError(str(self.db_systems))

        #insert system found in instance
        self.__insert_system(system)

        #get descriptors
        users_queue = await func.load_user_descriptors(
            system_id = self.id,
            db = self.db_users
        )

        #load descriptors
        self.detector.load_users(users = users_queue)

    def __get_cron_trigger(self) -> CronTrigger:
        """get crontrigger of now every day

        Returns:
            CronTrigger: trigger of action
        """
        _date = datetime.now(tz=get_localzone())
        return CronTrigger(hour=_date.hour, minute=_date.minute, second=_date.second, timezone=get_localzone())

    def __insert_system(self, system: dict) -> None:
        """Insert system inside actual instance

        Args:
            system (dict): values to be replaced in instance
        """
        self.id = system["id"]
        self.type_service = system["type_service"] if system["type_service"] is not None else self.type_service
        self.title = system["title"] if system["title"] is not None else self.title
        self.features = system["features"] if system["features"] is not None else self.features
        self.min_date_knowledge = system["min_date_knowledge"] if system["min_date_knowledge"] is not None else self.min_date_knowledge
        self.min_frequency = system["min_frequency"] if system["min_frequency"] is not None else self.min_frequency
        self.max_descriptor_distance = system["max_descriptor_distance"] if system["max_descriptor_distance"] is not None else self.max_descriptor_distance
        self.type_system = system["type_system"] if system["type_system"] is not None else self.type_system
        self.resize_factor = system["resize_factor"] if system["resize_factor"] is not None else self.resize_factor

    def set_conf(self, 
    receiver: 'function',
    detector: 'Detector',
    streamer: 'Streamer',
    stream_fps: float = 30,
    display_size: dict = None,
    date_format: str = None,
    draw_label: bool = None,
    cron_reload: CronTrigger = None,
    ) -> None:
        """Set configuration parameters for this instance .

        Args:
            receiver (function): provider of streaming video
            detector (Detector): face detector 
            streamer (Streamer): send result of streaming video
            stream_fps (float, optional): frame per second inside video. Defaults to 30.
            display_size (dict, optional): size of image process. Defaults to None.
        """
        self.receiver = receiver
        self.detector = detector
        self.streamer = streamer
        self.stream_fps = stream_fps

        #more info
        self.display_size = display_size if display_size is not None else self.display_size
        self.date_format = date_format if date_format is not None else self.date_format
        self.draw_label = draw_label if draw_label is not None else self.draw_label
        self.cron_reload = cron_reload if cron_reload is not None else self.cron_reload

    def get_obj(self) -> dict:
        """Get a dict of the attributes for this instance.

        Returns:
            dict: instance relevant features
        """
        return {
            "title": self.title,
            "type_service": self.type_service,
            "max_descriptor_distance": self.max_descriptor_distance,
            "min_date_knowledge": self.min_date_knowledge,
            "min_frequency": self.min_frequency,
            "features": self.features,
            "type_system": self.type_system,
            "resize_factor": self.resize_factor,
            "id": self.id,
        }

    async def start(self, cb: 'function' = None) -> None:
        
        await func.async_scheduler(job=self.__preload, trigger=self.cron_reload)
        
        while True: 
            _, img = next(self.receiver)
            if img is not None:
                img_processed = await self.process_unknows(
                    img = img,
                    resize_factor = self.resize_factor,
                    draw_label = self.draw_label,
                    features = self.features
                    )
                await self.streamer(img = img_processed, size = self.display_size, title = self.title)
                if cb is not None:
                    cb(img)
            
            if ft.ENVIROMENT == "DEV":
                sleep(5)

    async def process_unknows(self, img: np.array, resize_factor: float = 0.25, draw_label: bool = False, labels: tuple = ("Unknown", "Know"), features: list = []) -> 'np.array':
        """Process unknowns users.

        Args:
            img (np.array): img with faces
            resize_factor (float, optional): resize image to acelerate process. Defaults to 0.25.
            draw_label (bool, optional): draw labels over the faces. Defaults to False.
            labels (tuple, optional): labels for unknown users and know users. Defaults to ("Unknown" "Know").
            features (list, optional): list of features to save see utils.constants. Defaults to [].

        Returns:
            np.array: img
        """
        
        more_similar, less_similar = await self.detector.detect_unknowns(img, (1 - self.max_descriptor_distance), resize_factor, labels, features)

        # Display the results
        return_size = 1 / resize_factor

        for detection in less_similar:
                
            (top, right, bottom, left) = detection["location"]
            # Scale back up face locations since the frame we detected in was scaled to 1/4 size
            top *= return_size
            right *= return_size
            bottom *= return_size
            left *= return_size

            # Draw a box around the face
            cv2.rectangle(img, (left, top), (right, bottom), (0, 0, 255), 2)

            if draw_label is True:
                # Draw a label with a name below the face
                cv2.rectangle(img, (left, bottom - 35), (right, bottom), (0, 0, 255), cv2.FILLED)
                font = cv2.FONT_HERSHEY_DUPLEX
                cv2.putText(img, detection["name"], (left + 6, bottom - 6), font, 1.0, (255, 255, 255), 1)
        
            #save information of detection
            await self.set_detection({
                "system_id": self.id,
                "detection": detection["encoding"],
                "features": detection["features"],
                "author": str(uuid.uuid4()),
                "init_date": utils.get_actual_date(self.date_format),
                "last_date": utils.get_actual_date(self.date_format),
                "knowledge": False,
                "frequency": 0,
            })
        
        #evaluate knowed users
        for detection in more_similar:

            user = await func.find_user({
                "author": detection["author"]
            })
            if user is None:
                logging.error(f"Error user not found in BD, author: {detection['author']}")

            await self.evaluate_detection(user)

        return img

    async def evaluate_detection(self, user: dict) -> dict:
        """Evaluate the user s detection.

        Args:
            user (dict): user to be evaluated

        Returns:
            dict: user modified
        """
        prev_user = user
        diff_date = utils.get_date_diff_so_far(user.init_date, self.min_date_knowledge[1])

        if diff_date > self.min_date_knowledge[0] and user.frequency >= self.frequency:
            user["knowledge"] = True

        elif utils.get_date_diff_so_far(user.last_date, self.min_date_knowledge[1]) > 0:

            prev_days = utils.frequency(total = self.min_date_knowledge[0], percentage = 1, value = self.frequency, invert = True) 
            user.last_date = utils.get_actual_date(self.date_format)
            user.frequency = utils.frequency(self.min_date_knowledge[0], 1, prev_days + 1) 

        return await func.update_user({
            **user,
            "modified_on": utils.get_actual_date(self.date_format)
        }, self.db_users) if prev_user != user else user

    async def set_detection(self, user: dict) -> None:
        """Set the detection of a user in the database .

        Args:
            user (dict): [description]
        """
        try:    
            await func.insert_user(user, self.db_users)
        except Exception as e:
            logging.error(f"set_detection error to insert detection, {e}")



import asyncio
import threading
from flask import Flask, render_template
from mvfy.visual import VisualKnowledge
from mvfy.visual.utils import FlaskStreamer, Receiver, FaceRecognition
from mvfy.utils import constants as const

app = Flask(__name__)

features = [
    const.ALLOWED_FEATURES["AGE"],
    const.ALLOWED_FEATURES["GENDER"]
]
visual = VisualKnowledge(
    type_service=const.TYPE_SERVICE["LOCAL"],
    db_properties="mongodb://localhost:27017/",
    db_name="mvfy",
    max_descriptor_distance=0.7,
    min_date_knowledge=const.DAYS(7),
    type_system=const.TYPE_SYSTEM["OPTIMIZED"],
    features=features,
    title="mvfy_1"
)

visual.set_conf(
    detector=FaceRecognition(),
    receiver=Receiver.ip_cam_receiver(
        ip_cam="rtsp://mvfysystem:mvfysystem@192.168.1.6:8080/h264_ulaw.sdp"),
    streamer=Streamer.stream_local,
    display_size=(1080, 720)
)

streamer = FlaskStreamer(image_generator = visual)

@app.route("/")
def index():
    threading.Thread(target = streamer.start).start()
    return render_template(streamer.url_template)

@app.route("/stream")
async def index():
    return await streamer.get_frame()

if __name__ == "__main__":
    
    asyncio.run(streamer)



import asyncio
from asyncio import Queue, Task
from datetime import datetime
import logging
import uuid
from entities.visual_knowledge_entities import System
from data_access.visual_knowledge_db import SystemDB, UserDB
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger
from mvfy.use_cases.visual_knowledge_cases import SystemUseCases
from use_cases.visual_knowledge_cases import UserUseCases
from ..utils import index as utils

async def async_scheduler(job: 'function', trigger: CronTrigger, type: 'str' = "cron", *kargs) -> None:

    _scheduler = AsyncIOScheduler()
    _scheduler.add_job(job, 
    trigger=trigger if trigger is None else type, 
    id=str(uuid.uuid4()), *kargs)
    _scheduler.start()

    return _scheduler
    
def loop_manager(func: 'function') -> 'function':
    """Decorator for Manage Event Loop.

    Args:
        func ([type]): callback async function
    
    Returns:
        (function): result of func
    """
    async def wrapper_function(*args, **kargs):
        loop = asyncio.get_event_loop()
        if loop is None:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        return await func(*args, **kargs, loop=loop)
        
    return wrapper_function

async def async_queue_object_put(list: list[dict], keys: list[str], queue: 'Queue' = Queue()) -> None:
    """Generate Queue with object properties extracted

    Args:
        list (list[dict]): list of objects 
        keys (list[str]): list of pairs key:value, required
        queue (Queue, optional): [description]. Defaults to Queue().
    
    Example
        >>> ... await async_queue_object([..., {..., "system": 1}], ["system"])

    """
    for obj in list:
        await queue.put({
            f"{k}":v for k, v in obj.items() if k in keys
        })
    
    queue.task_done()

async def async_queue_object_get(queue: 'Queue', callback: 'function', args: tuple = ()) -> None:

    while not queue.empty():
        res = await queue.get()
        await callback(*args, queue_result=res)

    queue.task_done()

@loop_manager
async def load_user_descriptors(system_id: str, db: UserDB, loop: 'asyncio.AbstractEventLoop') -> 'utils.ThreadedGenerator|None':
    """Load user descriptors from database.

    Args:
        system_id (str): id of parent system
        db (str): data-base of users
        loop (asyncio.AbstractEventLoop): actual event loop

    Returns:
        list[dict]: [description]
    """

    use_cases = UserUseCases(db)

    results = await loop.run_in_executor(None, lambda: use_cases.get_users({
        "system_id": system_id
    }))

    if results == [] or results is None:
        return None

    users_queue = utils.ThreadedGenerator(results, daemon=True)
    users_queue.insert_action(cb=utils.extract_objects, args=(["detection", "author"]))

    return users_queue

@loop_manager
async def get_system(system: 'dict', db: SystemDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """Get information about a system.

    Returns:
        dict: system found
    """
    use_cases = SystemUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.get_system(system))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def insert_system(system: 'dict', db: SystemDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """add information about a system.

    Returns:
        str: id of system insert
    """
    use_cases = SystemUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.add_system(system))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def insert_user(user: 'dict', db: UserDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """add information about a user.

    Returns:
        str: id of user insert
    """
    use_cases = UserUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.add_user(user))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def find_user(filter: 'dict', db: UserDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """search information about a user.

    Returns:
        str: id of user insert
    """
    use_cases = UserUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.get_user(filter))

    if result == [] or result is None:
        return None

    return result


from typing import Iterable
import cv2
import socketio

class Receiver:
    def __init__(self) -> None:
        pass
    
    @staticmethod
    def ip_cam_receiver(ip_cam: 'str') -> Iterable:
        def inside_function():
            stream = None
            try:
                print(f"conecting.... {ip_cam}")
                if stream is None:
                    stream = cv2.VideoCapture(ip_cam)
                print("init the capture of image")

                if stream is None:
                    raise Exception("Stream error")

                while stream.isOpened():
                    try:
                        
                        yield stream.read()
                        
                        if stream is None:
                            print(f"conecting.... {ip_cam}")
                            stream = cv2.VideoCapture(ip_cam)
                    except Exception as e:
                        raise Exception(
                            f"Error in stream connection {e}")

            except Exception as e:
                raise Exception(
                    f"Error in connection to {ip_cam}, {e}")

        return inside_function()

    @staticmethod
    def socket_receiver(server_socket: 'tuple|str', buffer_size: float = 1024):
        def inside_function():
            sock = socketio.client()
            stream = None
            try:
                print(f"conecting.... {server_socket}")
                sock.connect(server_socket)
                print("init the capture of image")

                while True:
                    try:
                        # if stream is None:
                        #     stream = socketio.on('connect')(lambda: yield stream.read())
                        
                        print(f"conecting.... {ip_cam}")
                        if stream is None:
                            stream = cv2.VideoCapture(ip_cam)
                    except:
                        raise Exception(
                            f"Error in stream connection {e}")

            except Exception as e:
                raise Exception(
                    f"Error in conection to {ip_cam}, {e}")

        return inside_function


import os
import cv2
from typing import Optional, Tuple
from abc import ABC, abstractmethod
from flask import Response
from pydantic import BaseModel
from queue import Queue
from mvfy.visual.utils.streamer.errors import StreamTemplateNotFound
from visual.utils import constants

class Streamer(BaseModel, ABC):
    image_generator: ImageGenerator
    image: Optional[Array] = None
    images_queue: Queue = Queue(30)

    @abstractmethod
    def start(self) -> None:
        pass 
    
    @abstractmethod
    def get_frame(self) -> None:
        pass 

class FlaskStreamer(Streamer):

    resize: Optional[Tuple[int, int]]
    extension: Optional[str] = ".jpg"
    
    @property
    def url_template(self) -> str:

        dir_name: str = os.path.dirname(os.path.abspath(__file__))
        template_file: str = os.path.join(dir_name, "stream_flask_template.html")
        
        if not os.path.exists(template_file):
            raise StreamTemplateNotFound(path_file = template_file)
        
        return template_file
        
    async def start(self)-> None:

        while True:

            image = self.image_generator()

            if image is not None:

                if self.resize is not None:

                    image = cv2.resize(image, self.resize)

                flag, buffer = cv2.imencode(self.extension, image, [cv2.IMWRITE_JPEG_QUALITY, 80])
                
                if not flag:
                    self.image = buffer
                    await self.images_queue.put(self.image)
            
    async def get_frame(self) -> None:

        frame: bytearray = await self.images_queue.get()
        images_bytes: bytes = b'--frame\r\n' b'Content-Type: image/jpeg\r\n\r\n' + frame + b'\r\n'
        self.images_queue.task_done()

        return Response(images_bytes, mimetype="multipart/x-mixed-replace; boundary=frame")




from datetime import datetime
import inspect
import logging
from queue import Queue
from threading import Thread
from time import strptime
from typing import Iterable, Iterator

def get_actual_date(format: str) -> datetime:
    """Get the actual date from a given format.

    Args:
        format (str): valid format datetime

    Returns:
        datetime: formatted datetime
    """
    date = datetime.now()
    try:
        return datetime.strptime(date, format) if format is not None else date
    except Exception as e:
        logging.error(f"get_actual_date - Error to format datetime {e}")
        return date
        
def distribute_object(object_left: dict, object_right: dict) -> dict:
    """Insert object_right inside object_left
    
    Expected: same keys in both dicts
    
    Args:
        object_left (dict): object to be inserted
        object_right (dict): object to insert

    Raises:
        ValueError: [description]

    Returns:
        dict: objects merged
    """
    temp_object = {}

    if get_diff_list(object_left.keys(), object_right.keys()) is not None or get_diff_list(object_left.keys(), object_right.keys()) != []:
        raise ValueError("Invalid keys between objects")

    for key, item in object_right.items():
        temp_object[key] = object_left[key] if item is None else item

    return temp_object

def get_diff_list(lists: 'tuple(list)',   _type: 'str' = 'all') -> list:
    """Get difference between two list

    Args:
        lists (tuple): two list to be compared
        _type (str, optional): _type of get diff:

        all - get all list values different
        left - get only left different values
        right - get only right different values
        
        Defaults to 'all'.

    Raises:
        ValueError: Invalid size of lists, expected: __len__ 2
        ValueError: Invalid _type of lists

    Returns:
        list: difference
    """
    if len(lists) != 2:
        raise ValueError("Invalid size of lists, expected: __len__ 2")

    if not is_iterable(lists[0]) or not is_iterable(lists[1]):
        raise ValueError("Invalid _type of lists")

    diff = list(set(lists[0]) ^ set(lists[1]))

    if _type == "left":
        diff = [column for column in diff if column in lists[0]]
    
    elif _type == "right":
        diff = [column for column in diff if column in lists[1]]

    elif _type == "left":
        pass
    
    return diff

def is_iterable(posibleList):
    """Validate if element is iterable

    Args:
        posibleList (Any): posible iterable element

    Returns:
        bool: if element is iterable
    """
    try:
        if isinstance(posibleList, (tuple, list)) or hasattr(posibleList, "__iter__"):
            _ = posibleList[0]
            return True

        return False
    except Exception as e:
        return False

def extract_objects(object_list: list[dict], keys: list[str]) -> list[dict]:
    """Extracts keys from a list of objects.

    Args:
        object_list (list[dict]): list of objects
        keys (list[str]): keys that be extracted

    Returns:
        list[dict]: list of objects
    """
    res = []
    object_list = object_list if is_iterable(object_list) else [object_list]

    for obj in object_list:
        res.append({
            f"{k}":v for k, v in obj.items() if k in keys
        })

    return res

def get_date_diff_so_far(date: datetime, _type: str = "days") -> float:
    """Returns the difference between two date objects.

    Args:
        date (datetime): date to be compared with today
        _type (str, optional): type of difference <"days" or "weeks" or "months">. Defaults to "days".

    Returns:
        float: difference between "date" and now
    """
    date_now = get_actual_date()
    if date is not None:
        res = None
        if _type == "days":
            res = abs((date_now - date).days) 
        elif _type == "weeks":
            res = abs((date_now - date).days) // 7
        elif _type == "months":
            res = abs((date_now.year - date.year) * 12 + date_now.month - date.month)

        return res
    
def frequency(total: float, percentage: float, value: float, invert: bool = False) -> float:
    """Return frequency of a given value .

    Freq = v*t/p

    v-> value
    t-> total
    p-> percentage

    Args:
        total (float): total value
        percentage (float): percentage value between 0 and 1
        value (float): value
        invert (bool, optional): change total by percentage in formula. Defaults to False.

    Returns:
        float: frequency result
    """
    return (value * total) / percentage if invert is True else (value * percentage) / total

#------------------------------------------------------------------------------- CLASES ----------------------------------------------------------------------------------------

class ThreadedGenerator():
    """
    Generator that runs on a separate thread, returning values to calling
    thread. Care must be taken that the iterator does not mutate any shared
    variables referenced in the calling thread.
    """

    def __init__(self, 
        iterator: Iterator,
        daemon=False,
        Thread=Thread,
        sentinel: 'Any' = None,
        queue=Queue):

        self.size = len(iterator)
        self._iterator = iterator
        self._sentinel = sentinel
        self._queue = queue()
        self._thread = Thread(
            target=self._run
        )
        self._thread.daemon = daemon
        self._cb = None

    def insert_action(self, cb: 'function', args=()):
        """Add a callback to the intance.

        Args:
            cb (function): callback function
            args (tuple, optional): args. Defaults to ().

        """
        self._cb = ( lambda value: cb(value, *args) ) if args == () else ( lambda value: cb(value) )

    def __repr__(self):
        return 'ThreadedGenerator({!r})'.format(self._iterator)

    def _run(self):
        """Execute queue put process
        """
        for value in self._iterator:
            if self._cb is not None:
                try:
                    value = self._cb(value)
                except Exception as e:
                    logging.warning(f"{inspect.currentframe().f_code.co_name} Error inside function: \n {e}")

            self._queue.put(value)
            
        self._queue.put(self._sentinel)
        self._queue.task_done()

    def __iter__(self):
        """Iterate over the values in the queue .

        Yields:
            Any: value in the queue
        """
        self._thread.start()
        end = 0

        for value in iter(self._queue.get, self._sentinel):
            if value != self._sentinel:
                end += 1

            if end >= self.size:
                break

            yield value
        
        self._thread.join()




from dataclasses import dataclass
import logging

import cv2
import numpy as np
from cv2 import Mat
from abc import ABC, abstractmethod
from typing import Dict, Iterable, Optional, Tuple
from deepface import DeepFace
import face_recognition

from ....utils import index as utils

@dataclass
class Detector(ABC):

    authors: list = []
    encodings: list = []
    resize_factor: Optional[float] = 0.25

    @abstractmethod
    async def detect(self, image: Mat):
        pass

@dataclass
class DetectorUnknows(Detector):

    labels: Optional[tuple] = ("Unknown" "Know")
    min_descriptor_distance: Optional[float] = 0.6
    actual_img: Optional[np.array] = np.array([])

    async def detect(self, image: Mat) -> Tuple[Tuple[Dict], Tuple[Dict]]:
        """Detect unkwnows in image

        Args:
            image (Mat): image with faces to compare

        Returns:
            Tuple[Tuple[Dict], Tuple[Dict]]: more_similar, less_similar
        """
        self.actual_img = image

        self.redim_image()
        face_locations = face_recognition.face_locations(self.actual_img)
        face_encodings = []

        if not face_locations is None and face_locations != []:
            face_encodings = face_recognition.face_encodings(self.actual_img, face_locations)

        more_similar = []
        less_similar = []

        for idx, face_encoding in enumerate(face_encodings):
            face_distances = face_recognition.face_distance(self.encodings, face_encoding)
            if len(face_distances) > 0:
                best_match_index = np.argmin(face_distances)

                if face_distances[best_match_index] > self.min_descriptor_distance:
                    less_similar.append({
                        "name": self.labels[0],
                        "location": face_locations[idx],
                        "distance": face_distances[best_match_index],
                        "encoding": face_encoding,
                        "features": []
                    })
                else:
                    more_similar.append({
                        "name": self.labels[1],
                        "location": face_locations[idx],
                        "distance": face_distances[best_match_index],
                        "author": self.authors[best_match_index],
                        "encoding": face_encoding,
                        "features": []
                    })
            else:
                less_similar.append({
                    "name": self.labels[0],
                    "location": face_locations[idx],
                    "distance": 0,
                    "encoding": face_encoding,
                    "features": []
                })

        more_similar = utils.ThreadedGenerator(more_similar, daemon=True)
        less_similar = utils.ThreadedGenerator(less_similar, daemon=True)

        return more_similar, less_similar
        
    def redim_image(self) -> None:

        self.actual_img = cv2.resize(
            self.actual_img, 
            disize = (0, 0), 
            fx = self.resize_factor, 
            fy = self.resize_factor)

        self.actual_img = self.actual_img[:, :, ::-1]  # BGR to RBG
    
